#
# BTZen - Bluetooh Smart sensor reading library.
#
# Copyright (C) 2015 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .core import connect, Temperature, Humidity, Pressure, Light, \
    Accelerometer

__all__ = [
    'connect', 'Temperature', 'Humidity', 'Pressure', 'Light',
    'Motion',
]

# vim: sw=4:et:ai



#
# BTZen - Bluetooh Smart sensor reading library.
#
# Copyright (C) 2015 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Data conversion functions for various sensors.
"""

import struct

SHT21_TEMP = 175.72 / 65536
SHT21_HUMIDITY = 125 / 65536
HDC1000_HUMIDITY = 65536 / 100
MPU9250_GYRO = 65536 / 500
MPU9250_ACCEL_2G = 32768 / 2

BYTE_SHIFT = 1, 8, 16

to_int = lambda data: sum(v << b for v, b in zip(data, BYTE_SHIFT))

sht21_humidity = lambda data: -6.0 + SHT21_HUMIDITY * (to_int(data[2:]) & 0xfffc)
sht21_temp = lambda data: -46.85 + SHT21_TEMP * to_int(data[:2])
tmp006_temp = lambda data: to_int(data[2:]) / 128.0
bmp280_pressure = lambda data: to_int(data[3:])
hdc1000_humidity = lambda data: to_int(data[2:]) / HDC1000_HUMIDITY

def opt3001_light(data):
    v = to_int(data)
    m = (v & 0x0FFF) / 100
    e = (v & 0xF000) >> 12
    return m * (2 << e)


def epcos_t5400_pressure(calib, data):
    temp, pressure = struct.unpack('<hH', bytearray(data))

    c = calib
    t2 = temp << 2
    sens = c[2] + ((c[3] * temp) >> 17) + ((c[4] * t2) >> 34)
    off = (c[5] << 14) + ((c[6] * temp) >> 3) + ((c[7] * t2) >> 19)
    return (sens * pressure + off) >> 14


def mpu9250_motion(data):
    data = struct.unpack('<9h', bytes(data))
    #gyro = tuple(v / MPU9250_GYRO for v in data[:3])
    accel = tuple(v / MPU9250_ACCEL_2G for v in data[3:6])
    #magnet = data[6:]
    return accel

# vim: sw=4:et:ai



#
# BTZen - Bluetooh Smart sensor reading library.
#
# Copyright (C) 2015 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
The identificators for specific sensors can be found at

CC2541DK
    http://processors.wiki.ti.com/index.php/SensorTag_User_Guide
CC2650STK
    http://processors.wiki.ti.com/index.php/CC2650_SensorTag_User's_Guide
"""

import asyncio
import functools
import logging
import struct
from collections import namedtuple

from . import dbus
from .import conv

logger = logging.getLogger(__name__)


Parameters = namedtuple('Parameters', [
    'uuid_dev', 'uuid_conf', 'uuid_period', 'config_on', 'config_off',
])


def converter_epcos_t5400_pressure(dev, p_conf):
    p_calib = dbus.find_sensor(dev, dev_uuid(0xaa43))
    p_conf._obj.WriteValue([2])
    calib = p_calib._obj.ReadValue()
    calib = struct.unpack('<4H4h', bytearray(calib))
    return functools.partial(conv.epcos_t5400_pressure, calib)


dev_uuid = 'f000{:04x}-0451-4000-b000-000000000000'.format


# (sensor name, sensor id): data converter
DATA_CONVERTER = {
    ('TI BLE Sensor Tag', dev_uuid(0xaa01)):lambda *args: conv.tmp006_temp,
    ('TI BLE Sensor Tag', dev_uuid(0xaa21)): lambda *args: conv.sht21_humidity,
    ('TI BLE Sensor Tag', dev_uuid(0xaa41)): converter_epcos_t5400_pressure,
    ('SensorTag 2.0', dev_uuid(0xaa01)):lambda *args: conv.tmp006_temp,
    ('SensorTag 2.0', dev_uuid(0xaa21)): lambda *args: conv.hdc1000_humidity,
    ('SensorTag 2.0', dev_uuid(0xaa41)): lambda *args: conv.bmp280_pressure,
    ('SensorTag 2.0', dev_uuid(0xaa71)): lambda *args: conv.opt3001_light,
    ('SensorTag 2.0', dev_uuid(0xaa81)): lambda *args: conv.mpu9250_motion,
}

data_converter = lambda device, dev_id: \
    DATA_CONVERTER[(str(device.Name), dev_id)]


def connect(bus, mac):
    """
    Connect to device with MAC address `mac`.

    :param mac: Bluetooth device MAC address.
    """
    device = dbus.get_device(bus, mac)
    return device


class Reader:
    def __init__(self, bus, device, params, loop=None):
        super().__init__()
        self._loop = asyncio.get_event_loop() if loop is None else loop

        self._params = params
        self._device = self._find(bus, device, params.uuid_dev)
        self._dev_conf = self._find(bus, device, params.uuid_conf)
        self._dev_period = self._find(bus, device, params.uuid_period)

        self._values = asyncio.Queue()

        factory = data_converter(device, params.uuid_dev)
        self._converter = factory(device, self._dev_conf)

        self.set_interval(1)
        self._dev_conf._obj.WriteValue(params.config_on)


    def _find(self, bus, device, uuid):
        obj = dbus.find_sensor(bus, device, uuid)
        if obj is None:
            raise ValueError(
                'Cannot find object for uuid {}'.format(uuid)
            )
        if __debug__:
            logger.debug('object for uuid {} found'.format(uuid))

        return obj


    def set_interval(self, interval):
        self._dev_period._obj.WriteValue([interval * 100])


    def read(self):
        """
        Read data from sensor.
        """
        value = self.device._obj.ReadValue()
        return self._converter(value)


    async def read_async(self):
        """
        Read data from sensor in asynchronous manner.

        This method is a coroutine.
        """
        def cb(value):
            value = self._converter(value)
            self._loop.call_soon_threadsafe(self._values.put_nowait, value)

        def error_cb(*args):
            raise TypeError(self.__class__.__name__, *args) # FIXME

        self._device._obj.ReadValue(reply_handler=cb, error_handler=error_cb)
        value = await self._values.get()
        return value


    def close(self):
        self._dev_conf._obj.WriteValue(self._params.config_off)
        logger.info('{} device closed'.format(self.__class__.__name__))



class Temperature(Reader):
    def __init__(self, bus, device, loop=None):
        dev = Parameters(
            dev_uuid(0xaa01),
            dev_uuid(0xaa02),
            dev_uuid(0xaa03),
            [1],
            [0],
        )
        super().__init__(bus, device, dev, loop=loop)


class Humidity(Reader):
    def __init__(self, bus, device, loop=None):
        dev = Parameters(
            dev_uuid(0xaa21),
            dev_uuid(0xaa22),
            dev_uuid(0xaa23),
            [1],
            [0],
        )
        super().__init__(bus, device, dev, loop=loop)



class Pressure(Reader):
    def __init__(self, bus, device, loop=None):
        dev = Parameters(
            dev_uuid(0xaa41),
            dev_uuid(0xaa42),
            dev_uuid(0xaa44),
            [1],
            [0],
        )
        super().__init__(bus, device, dev, loop=loop)


class Light(Reader):
    def __init__(self, bus, device, loop=None):
        dev = Parameters(
            dev_uuid(0xaa71),
            dev_uuid(0xaa72),
            dev_uuid(0xaa73),
            [1],
            [0],
        )
        super().__init__(bus, device, dev, loop=loop)


class Accelerometer(Reader):
    ACCEL_X = 0x20
    ACCEL_Y = 0x10
    ACCEL_Z = 0x08

    def __init__(self, bus, device, loop=None):
        config = self.ACCEL_X | self.ACCEL_Y | self.ACCEL_Z
        dev = Parameters(
            dev_uuid(0xaa81),
            dev_uuid(0xaa82),
            dev_uuid(0xaa83),
            struct.pack('<H', config),
            b'\x00\x00',
        )
        super().__init__(bus, device, dev, loop=loop)


# vim: sw=4:et:ai



#!/usr/bin/env python3

#
# Copyright (C) 2015 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""
Build setup for btsensor library.
"""

from setuptools import setup, find_packages

setup(
    name='btzen',
    version='0.0.8',
    author='Artur Wroblewski',
    author_email='wrobell@pld-linux.org',
    url='https://github.com/wrobell/btzen',
    description='BTZen - Bluetooh Smart sensor reading library',
    setup_requires = ['setuptools_git >= 1.0',],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages('.'),
    scripts=('bin/btzen',),
    include_package_data=True,
)

# vim: sw=4:et:ai




#
# BTZen - Bluetooh Smart sensor reading library.
#
# Copyright (C) 2015 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
D-Bus support classes and functions.
"""

import dbus


class Proxy:
    """
    D-Bus object proxy to expose properties via simple attribute access.
    """
    def __init__(self, obj, iface):
        self._obj = obj
        self._iface = iface
        self._properties = dbus.Interface(obj, 'org.freedesktop.DBus.Properties')


    def __getattr__(self, name):
        return self._properties.Get(self._iface, name)
        

def get_device(bus, mac):
    bus_name = 'org.bluez'
    path = '/org/bluez/hci0/dev_{}'.format(mac.replace(':', '_'))
    proxy = bus.get_object(bus_name, path)
    device = dbus.Interface(proxy, dbus_interface='org.bluez.Device1')
    return Proxy(device, 'org.bluez.Device1')


def load_object(bus, path, iface):
    proxy = bus.get_object('org.bluez', path)
    return Proxy(dbus.Interface(proxy, iface), iface)


def load_objects(bus, paths, iface):
    return (load_object(bus, p, iface) for p in paths)


def get_services(bus, device):
    return load_objects(bus, device.GattServices, 'org.bluez.GattService1')


def get_characteristics(bus, service):
    return load_objects(bus, service.Characteristics, 'org.bluez.GattCharacteristic1')


def get_descriptors(bus, characteristics):
    return load_objects(bus, characteristics.Descriptors, 'org.bluez.GattDescriptor1')


def find_sensor(bus, device, uuid):
    items = (c for s in get_services(bus, device) for c in get_characteristics(bus, s))
    return next((c for c in items if c.UUID == uuid), None)


# vim: sw=4:et:ai
